from dataclasses import dataclass, field
from dbt.adapters.base.relation import BaseRelation, Policy

from typing import Optional, TypeVar, Type
from dbt.contracts.relation import RelationType


@dataclass
class IBMDB2QuotePolicy(Policy):
    database: bool = False
    schema: bool = False
    identifier: bool = False


@dataclass
class IBMDB2IncludePolicy(Policy):
    database: bool = False
    schema: bool = True
    identifier: bool = True


@dataclass(frozen=True, eq=False, repr=False)
class IBMDB2Relation(BaseRelation):
    quote_policy: IBMDB2QuotePolicy = field(default_factory=lambda: IBMDB2QuotePolicy())
    include_policy: IBMDB2IncludePolicy = field(default_factory=lambda: IBMDB2IncludePolicy())

    @staticmethod
    def add_ephemeral_prefix(name: str):
        return f'DBT_CTE__{name}'



from dbt.adapters.ibmdb2.connections import IBMDB2ConnectionManager
from dbt.adapters.ibmdb2.connections import IBMDB2Credentials
from dbt.adapters.ibmdb2.impl import IBMDB2Adapter

from dbt.adapters.base import AdapterPlugin
from dbt.include import ibmdb2


Plugin = AdapterPlugin(
    adapter=IBMDB2Adapter,
    credentials=IBMDB2Credentials,
    include_path=ibmdb2.PACKAGE_PATH)



from contextlib import contextmanager

from dataclasses import dataclass

import dbt.exceptions
from dbt.adapters.base import Credentials
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.connection import Connection
from dbt.adapters.sql import SQLConnectionManager
from dbt.logger import GLOBAL_LOGGER as logger

from typing import (
    Type,
    Iterable,
    Optional
)

import ibm_db
import ibm_db_dbi

@dataclass
class IBMDB2Credentials(Credentials):
    host: str
    database: str
    schema: str
    user: str
    password: str
    port: int = 50000
    protocol: str = 'TCPIP'
    extra_connect_opts: Optional[str] = None

    @property
    def type(self):
        return 'ibmdb2'

    def _connection_keys(self):
        return ('host', 'database', 'schema', 'user', 'password', 'port', 'protocol', 'extra_connect_opts')


class IBMDB2ConnectionManager(SQLConnectionManager):
    TYPE = 'ibmdb2'

    @contextmanager
    def exception_handler(self, sql: str):
        try:
            yield
        except ibm_db_dbi.DatabaseError as exc:
            self.release()
            logger.debug('ibm_db_dbi error: {}'.format(str(exc)))
            logger.debug("Error running SQL: {}".format(sql))
            raise dbt.exceptions.DbtDatabaseError(str(exc))
        except Exception as exc:
            self.release()
            logger.debug("Error running SQL: {}".format(sql))
            logger.debug("Rolling back transaction.")
            raise dbt.exceptions.DbtRuntimeError(str(exc))

    @classmethod
    def open(cls, connection):
        if connection.state == 'open':
            logger.debug('Connection is already open, skipping open.')
            return connection

        credentials = connection.credentials

        def connect():
            con_str = f"DATABASE={credentials.database}"
            con_str += f";HOSTNAME={credentials.host}"
            con_str += f";PORT={credentials.port}"
            con_str += f";PROTOCOL={credentials.protocol}"
            con_str += f";UID={credentials.user}"
            con_str += f";PWD={credentials.password}"

            if credentials.extra_connect_opts is not None and credentials.extra_connect_opts != "":
                con_str += f";{credentials.extra_connect_opts}"

            handle = ibm_db_dbi.connect(con_str, '', '')
            handle.set_autocommit(False)

            return handle

        retryable_exceptions = [ibm_db_dbi.OperationalError,]

        return cls.retry_connection(
            connection,
            connect=connect,
            logger=logger,
            retry_limit=3,
            retry_timeout=5,
            retryable_exceptions=retryable_exceptions
        )

    @classmethod
    def cancel(self, connection):
        connection_name = connection.name

        logger.info("Cancelling query '{}' ".format(connection_name))

        try:
            connection.handle.close()
        except Exception as e:
            logger.error('Error closing connection for cancel request')
            raise Exception(str(e))

    @classmethod
    def get_credentials(cls, credentials):
        return credentials

    @classmethod
    def get_response(cls, cursor) -> AdapterResponse:

        message = 'OK'
        rows = cursor.rowcount

        return AdapterResponse(
            _message=message,
            rows_affected=rows
        )

    def add_begin_query(self):
        pass



from dbt.adapters.sql import SQLAdapter
from dbt.adapters.ibmdb2 import IBMDB2ConnectionManager
from dbt.adapters.ibmdb2.relation import IBMDB2Relation
from dbt.adapters.ibmdb2.column import IBMDB2Column

from typing import Dict
from dbt.utils import filter_null_values
from dbt.exceptions import CompilationError
from dbt.adapters.base.meta import available
from typing import Optional

import agate


class IBMDB2Adapter(SQLAdapter):

    ConnectionManager = IBMDB2ConnectionManager
    Relation = IBMDB2Relation
    Column = IBMDB2Column

    @classmethod
    def date_function(cls):
        return 'CURRENT_DATE'

    def is_cancelable(cls):
        return False

    # Notes: DB2 always need a FROM
    def debug_query(self):
        self.execute('select 1 as one from sysibm.sysdummy1')

    # DB2 Notes:
    # ----------
    # To allow for none quoting and keep lowercase naming of schema (in profiles)
    # and identifiers (models via filenames) we need to modify the search params,
    # otherwise dbt will search for lowercase and find uppercase which will lead
    # to exceptions.
    def _make_match_kwargs(self, database: str, schema: str, identifier: str) -> Dict[str, str]:
        quoting = self.config.quoting

        if identifier is not None and quoting["identifier"] is False:
            identifier = identifier.upper()

        if schema is not None and quoting["schema"] is False:
            schema = schema.upper()

        # DB2 Notes:
        # ----------
        # No need to change database to upper. Then database is always created
        # before hand and will always have the same value as in the profiles
        # config because we set it 'manually' in list_relations_without_caching.

        #if database is not None and quoting["database"] is False:
        #    database = database.upper()

        return filter_null_values(
            {
                "database": database,
                "identifier": identifier,
                "schema": schema,
            }
        )

    @available
    def quote_seed_column(self, column: str, quote_config: Optional[bool]) -> str:
        quote_columns: bool = False
        if isinstance(quote_config, bool):
            quote_columns = quote_config
        elif quote_config is None:
            pass
        else:
            raise CompilationError(
                f'The seed configuration value of "quote_columns" has an '
                f"invalid type {type(quote_config)}"
            )

        if quote_columns:
            return self.quote(column)
        else:
            return column

    @classmethod
    def convert_text_type(cls, agate_table, col_idx):
        column = agate_table.columns[col_idx]
        lens = (len(d.encode("utf-8")) for d in column.values_without_nulls())
        max_len = max(lens) if lens else 64
        length = max_len if max_len > 16 else 16
        return "varchar({})".format(length)

    @classmethod
    def convert_date_type(cls, agate_table, col_idx):
        return "timestamp"

    @classmethod
    def convert_datetime_type(cls, agate_table, col_idx):
        return "timestamp"

    @classmethod
    def convert_boolean_type(cls, agate_table, col_idx):
        return "decimal(1)"

    @classmethod
    def convert_number_type(cls, agate_table, col_idx):
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))
        return "decimal"

    @classmethod
    def convert_time_type(cls, agate_table, col_idx):
        return "timestamp"

    def valid_incremental_strategies(self):
        """The set of standard builtin strategies which this adapter supports out-of-the-box.
        Not used to validate custom strategies defined by end users.
        """
        # >>> DB2 Note:
        # "delete+insert" and "insert_overwrite" not supported yet
        # <<<
        return ["append", "merge"]


from dataclasses import dataclass
from typing import Dict, ClassVar

from dbt.adapters.base.column import Column


@dataclass
class IBMDB2Column(Column):

    TYPE_LABELS: ClassVar[Dict[str, str]] = {
        "STRING": "VARCHAR(4000)",
        "TIMESTAMP": "TIMESTAMP",
        "FLOAT": "DECFLOAT",
        "INTEGER": "INTEGER",
    }

    STRING_DATATYPES = {'char', 'nchar', 'varchar', 'nvarchar'}
    NUMBER_DATATYPES = {'decimal', 'decfloat'}

    @property
    def data_type(self) -> str:
        if self.is_string():
            return self.ibmdb2_string_type(self.dtype, self.string_size())
        elif self.is_numeric():
            return self.numeric_type(self.dtype, self.numeric_precision, self.numeric_scale)
        else:
            return self.dtype

    @classmethod
    def ibmdb2_string_type(cls, dtype: str, size: int = None):
        """
            - CHAR(SIZE)
            - VARCHAR(SIZE)
            - NCHAR(SIZE) or NCHAR
            - NVARCHAR(SIZE)
        """
        if size is None:
            return dtype
        else:
            return "{}({})".format(dtype, size)

    def is_numeric(self) -> bool:
        if self.dtype.lower() in self.NUMBER_DATATYPES:
            return True
        return super().is_numeric()

    def is_string(self) -> bool:
        if self.dtype.lower() in self.STRING_DATATYPES:
            return True
        return super().is_string()
